/*
 *  FILEDATE        *** For Wildcat v2.00 files only! ***
 *
 *  Description -   A program to make a formatted listing of the records
 *                  that are found in the FILESPEC.DAT file on a Wildcat
 *                  BBS v2.00.  Prompted by the desire to continue making
 *                  a daily-updated list available as the "newsletter"
 *                  selection on my BBS, now that the data formats have
 *                  been changed.  (The same file is also archived for
 *                  downloading.)
 *                      FILEDATE creates up to 26 linked lists, which will
 *                  each grow as large as necessary to hold your file
 *                  information.  There is no need for it to know ahead of
 *                  time how many files the directories contain.  A listing
 *                  is created in file-date order, with categories listed
 *                  in the order they appear in CONFIGWC.DAT.
 *                       An additional benefit...   if you've never
 *                  converted Turbo Pascal strings into standard ASCIZ
 *                  strings, the function at the end of this source will
 *                  show you how.  It works like strcpy(), but converts
 *                  from Pascal to C along the way.
 *                       Also, there's a function to convert the WC date
 *                  values (stored as number of days since 01/01/1600!)
 *                  into strings that have the common format.
 *
 *  Usage -         Just run:
 *                              FILEDATE filename
 *
 *                  No other parameters, no required environment
 *                  variables, no complications.  Make sure that this
 *                  program is in a directory within your DOS path, and
 *                  that the file named CONFIGWC.BBS is in the current
 *                  default directory.
 *                      If you prefer, you can call this program from
 *                  within a batch file (eg. as an event that runs each
 *                  night), giving you an opportunity to CD around.
 *                  FILEDATE creates a file of the name given in its
 *                  command-line argument.
 *
 *  Return -        Returns an errorlevel of (1) for any error.  Returns
 *                  errorlevel (0) for normal completion.
 *
 *  Copyright -     This source file and the enclosed executable file are
 *
 *                             Copyright Jerry Houston, 1988
 *                            COLLEGE CORNER BBS, Bellevue, WA
 *                                   (206) 643-0804
 *
 *                  You may do anything with them that you want, except
 *                  sell them.  Please distribute freely, alone or in
 *                  connection with any other product.
 *
 *  Warrenty -      The program FILEDATE.EXE and the file FILEDATE.C are
 *                  provided free of charge, without any warrenty at all.
 *                  All risk becomes that of the user.
 *
 *  Version -       1.0  -  10 April 1988  -  Jerry Houston
 *                  2.0  -  28 February 1989 - Jerry Houston
 *                  2.1  -  2 July 1989 - Jerry Houston, changed to
 *                          work with WC v2.0 BBS
 *
 */
#include <MALLOC.H>
#include <STDIO.H>
#include <TIME.H>
#include <STDLIB.H>
#include <STRING.H>

#define FALSE 0
#define TRUE !FALSE
#define NAMESIZE    31L      /* maximum length of the names in Pascal or C */
#define NAMESTART  4209L     /* starting position in config file for names */

#pragma pack (1)

typedef struct rec1
{
    long    lBtreeStatus;
    char    szFileName[13],
	    szPassword[15];
    long    lFileDate,
	    lFileTime,
	    lLastDate,
	    lLastTime;
    char    szDesc1[41],
	    szDesc2[41],
	    szUploadedBy[38],
	    cFileArea;
    long    lNumAccess,
	    lSize;
    char    iPasswordNeeded;
}
    FileRecord;


/*  Each node in the list will contain these fields, copied from the
 *  record above, which is read from the FILESPEC.DAT file.
 */
typedef struct rec2
{
    char    szFileName[13];
    long    lFileDate,
	    lFileTime;
    char    szDesc1[41],
	    szDesc2[41];
    long    lNumAccess,
	    lSize;
    struct rec2  *pnFlink,
                 *pnBlink;
}
    NodeRecord;


/*  This record forms the array of names and pointers to linked lists.
 */
typedef struct rec3
{
    char    szFileConf[31];
    NodeRecord  *pnListPtr;
}
    ConfArray;


/*  Some static variables that are used to read the system time and date,
 *  and convert them to an ASCII string for the output file listing.
 */
struct tm   *ptmNewTime;
time_t      aclock;


/*  Prototype for the function that converts a Pascal string to a C
 *  string.
 */
char  *strpas2c(char *, char *);

/*  Prototype for the function that converts a long to a string date.
 *  The long contains the number of days since 01/01/1600.
 */
char *long2date(long);


int main(int argc, char **argv)
{
    static ConfArray   PtrArray[26];
           FileRecord  InBuff;
           NodeRecord  *pnCurrent,
                       *pnNewNode;

    int  iNameCount,
         iCharCount,
         iIndex,
         iArrayIndex;

    FILE *pfConfigFile,
         *pfListFile,
	 *pfDataFile;

    char szFileSpecPath[67];



    if( argc < 2 )
    {
	fprintf( stderr, "\n\nUSAGE:   C:\> FILEDATE  bulletin_file\n\n" );
        exit( 0 );
    }

    fprintf(stderr, "\n\nFILEDATE:  Reading file data, please wait...");
    time(&aclock);
    ptmNewTime = localtime(&aclock);

/*  open CONFIGWC.BBS file to read the categories
 */
    pfConfigFile = fopen("configwc.bbs", "rb");
    if (! pfConfigFile)
    {
	fprintf(stderr, "\n\nFILEDATE: Cannot locate "
				   "CONFIGWC.BBS file for input.\n\n");
        exit (1);
    }


    fseek(pfConfigFile, 747L, SEEK_SET);
    iCharCount = fgetc(pfConfigFile);
    for (iIndex = 0; iIndex < iCharCount; ++iIndex)
	szFileSpecPath[iIndex] = (char)fgetc(pfConfigFile);
    szFileSpecPath[iIndex] = '\0';
    strcat(szFileSpecPath, "FILESPEC.DAT" );

/*  Create 26 starting list nodes, and assign their pointers to the
 *  elements of the array.  Will assign conference names to the elements
 *  in the next section.  Kept separate for code simplicity.
 */
    for (iIndex = 0; iIndex < 26; ++iIndex)
    {
        PtrArray[iIndex].pnListPtr = (NodeRecord *) malloc(sizeof(NodeRecord));
        if (PtrArray[iIndex].pnListPtr == NULL)
        {
	    fprintf(stderr,"\n\nFILEDATE: Cannot allocate enough memory.\n\n");
            exit(1);
        }

        PtrArray[iIndex].pnListPtr->pnFlink = NULL;
        PtrArray[iIndex].pnListPtr->pnBlink = PtrArray[iIndex].pnListPtr;
	PtrArray[iIndex].pnListPtr->szFileName[0] = NULL;
	PtrArray[iIndex].pnListPtr->lFileDate = 0xFFFF;
    }


/*  For up to 26 conference names, read the string length (first byte)
 *  then copy that many characters to the array of names.  Empty Pascal
 *  strings start with a NULL byte, making them empty C strings, too.
 */
    for (iNameCount = 0; iNameCount < 26; ++iNameCount)
    {
        fseek(pfConfigFile, (long)(NAMESTART+iNameCount*NAMESIZE), SEEK_SET);
        iCharCount = fgetc(pfConfigFile);
        for (iIndex = 0; iIndex < iCharCount; ++iIndex)
	    PtrArray[iNameCount].szFileConf[iIndex] = (char)fgetc(pfConfigFile);
	PtrArray[iNameCount].szFileConf[iIndex] = '\0';
    }


/*  Now the array contains 26 names associated with 26 pointers to nodes,
 *  which each point to a dummy (empty) initial node.  Done with this file.
 */
    fclose(pfConfigFile);


/*  open FILESPEC.DAT file to read file data
 */
    pfDataFile = fopen(szFileSpecPath, "rb");
    if (! pfDataFile)
    {
	fprintf(stderr, "\n\nFILEDATE: Cannot locate "
				   "%s file for ipnut.\n\n", strupr(szFileSpecPath));
        exit (1);
    }


/*  Next job is to read through the files database, and create a linked-
 *  list node for each of the records found there.  Each node will contain
 *  a filename, date, and two descriptions, plus a pointer to the next
 *  node in the same list.  Each node will be put into the linked list
 *  that pertains to that category.
 */
    fseek(pfDataFile, 178L, SEEK_SET);
    fread((char *)&InBuff, sizeof(FileRecord), 1, pfDataFile);

    while (! feof(pfDataFile))
    {


	iArrayIndex = InBuff.cFileArea - 65;
        if (iArrayIndex < 26 && iArrayIndex >=0)
        {
            pnNewNode = (NodeRecord *) malloc(sizeof(NodeRecord));
            if (pnNewNode == NULL)
            {
		fprintf(stderr,"\n\nFILEDATE: Cannot allocate "
                                                "enough memory.\n\n");
                exit(1);
            }

	    strpas2c(pnNewNode->szFileName,InBuff.szFileName);
            pnNewNode->lSize = InBuff.lSize;
	    pnNewNode->lNumAccess = InBuff.lNumAccess;
	    pnNewNode->lFileDate = InBuff.lFileDate;
	    if(pnNewNode->lFileDate < 139151L)
		pnNewNode->lFileDate = 139150L;
	    strpas2c(pnNewNode->szDesc1,InBuff.szDesc1);
	    strpas2c(pnNewNode->szDesc2,InBuff.szDesc2);

            pnCurrent = PtrArray[iArrayIndex].pnListPtr;
	while( pnCurrent->pnFlink && pnNewNode->lFileDate  > pnCurrent->lFileDate)
	      pnCurrent = pnCurrent->pnFlink;
	if( pnCurrent->pnFlink || pnNewNode->lFileDate < pnCurrent->lFileDate)

	   {
                pnCurrent = pnCurrent->pnBlink;

                pnCurrent->pnFlink->pnBlink = pnNewNode;
                pnNewNode->pnFlink = pnCurrent->pnFlink;
                pnCurrent->pnFlink = pnNewNode;
                pnNewNode->pnBlink = pnCurrent;
            }
            else
            {
                pnCurrent->pnFlink = pnNewNode;
                pnNewNode->pnBlink = pnCurrent;
                pnNewNode->pnFlink = NULL;
            }
        }
        fread((char *)&InBuff, sizeof(FileRecord), 1, pfDataFile);
    }

/* open file for output
 */
    pfListFile = fopen( argv[1], "wt");
    if (! pfListFile)
    {
	fprintf(stderr, "\n\nFILEDATE: Cannot open "
                                        "%s for output.\n\n", strupr(argv[1]) );
    }

/*  Last task is to do something with the lists.  Will format them to
 *  an output file such that all the records for one directory are listed
 *  together.  Simply a matter of traversing the array, traversing each
 *  list along the way.
 */
    for (iArrayIndex = 0; iArrayIndex < 26; ++iArrayIndex)
    {
        if (PtrArray[iArrayIndex].pnListPtr->pnFlink == NULL)
            continue;
	fprintf(pfListFile, "\n\n\n     Area %c: %s      Created: %s ",
            iArrayIndex+65,
	    PtrArray[iArrayIndex].szFileConf,
            asctime(ptmNewTime));


        pnCurrent = PtrArray[iArrayIndex].pnListPtr->pnFlink;
        while(pnCurrent)
        {
	    fprintf(pfListFile, "\n        %-12s  Size: %7ld    Date: %s    Times D/L: %ld",
		pnCurrent->szFileName,
                pnCurrent->lSize,
		long2date(pnCurrent->lFileDate),
		pnCurrent->lNumAccess);
            fprintf(pfListFile, "\n            %s",
		pnCurrent->szDesc1);
	    if(*(pnCurrent->szDesc2))
                fprintf(pfListFile, "\n            %s",
		    pnCurrent->szDesc2);
            fprintf(pfListFile, "\n");
            pnCurrent = pnCurrent->pnFlink;
        }
    }
    fclose(pfListFile);
    fprintf(stderr,"\n\n              ... finished!\n\n");
    return(0);
}



/*  strpas2c()  -  String from Pascal to C format.  Accepts address where
 *  a Pascal (ASCIC type) string has been copied, converts it to a C string
 *  (ASCIZ type).  Returns pointer to the C string, so this function can
 *  be used in an expression that requires a string.  Copies from the
 *  second argument to the first, just like strcat(), strcpy(), etc.
 */
char  *strpas2c(char  *szCString, char  *szPasString)
{
    int iIndex;

    for (iIndex = 0; iIndex < *(szPasString); ++iIndex)
	szCString[iIndex] = szPasString[iIndex+1];
    szCString[iIndex] = '\0';

    return(szCString);
}



/*  long2date()  -  Converts a long to a string date, where the long
 *                  contains the number of days since Jan 1, 1600.
 */
char *long2date(long lLongDate)
{
    static char szDateString[9];
    int  iNumberOfLeaps, iIsLeap, iDayInYear;
    int  iYear, iMonth, iDay;

    if( lLongDate < 139151L )     // if the date is bogus,
    {                             // return a **/**/** string
	szDateString[0] = '*';
	szDateString[1] = '*';
	szDateString[2] = '/';
	szDateString[3] = '*';
	szDateString[4] = '*';
	szDateString[5] = '/';
	szDateString[6] = '*';
	szDateString[7] = '*';
	szDateString[8] = '\0';
	return( szDateString );
    }

    iNumberOfLeaps = lLongDate / 1461L - 4L;
    lLongDate -= iNumberOfLeaps;
    iYear = lLongDate / 365 + 1600;
    if( !(iYear % 4) )
	iIsLeap = TRUE;
    else
	iIsLeap = FALSE;

    iDayInYear = lLongDate % 365;
    if(iIsLeap && iDayInYear > 60)
	--iDayInYear;

    if(iDayInYear < 32)
    {
      iMonth = 1;
      iDay = iDayInYear;
    }
    else
    {
      if (iDayInYear < 60)
      {
	iMonth = 2;
	iDay = iDayInYear - 31;
      }
      else
      {
	if( iDayInYear == 60 && iIsLeap )
	{
	  iMonth = 2;
	  iDay = 29;
	}
	else
	{
	  if( iDayInYear < 91 )
	  {
	    iMonth = 3;
	    iDay = iDayInYear - 59;
	  }
	  else
	  {
	    if( iDayInYear < 121 )
	    {
	      iMonth = 4;
	      iDay = iDayInYear - 90;
	    }
	    else
	    {
	      if(iDayInYear < 152)
	      {
		iMonth = 5;
		iDay = iDayInYear - 120;
	      }
	      else
	      {
		if(iDayInYear < 182)
		{
		  iMonth = 6;
		  iDay = iDayInYear - 151;
		}
		else
		{
		  if(iDayInYear < 213)
		  {
		    iMonth = 7;
		    iDay = iDayInYear - 181;
		  }
		  else
		  {
		    if(iDayInYear < 244)
		    {
		      iMonth = 8;
		      iDay = iDayInYear - 212;
		    }
		    else
		    {
		      if(iDayInYear < 274)
		      {
			iMonth = 9;
			iDay = iDayInYear - 243;
		      }
		      else
		      {
			if(iDayInYear < 305)
			{
			  iMonth = 10;
			  iDay = iDayInYear - 273;
			}
			else
			{
			  if(iDayInYear < 335)
			  {
			    iMonth = 11;
			    iDay = iDayInYear - 304;
			  }
			  else
			  {
			    iMonth = 12;
			    iDay = iDayInYear - 334;
			  }
			}
		      }
		    }
		  }
		}
	      }
	    }
	  }
	}
      }
    }

    sprintf( szDateString, "%02d/%02d/%02d", iMonth, iDay, iYear - 1900);
    return (szDateString);
}
